import os
import pickle
import tempfile
import time
from itertools import combinations_with_replacement as cwr

import healpy as hp
import numpy as np
import yaml
from pspy import pspy_utils, so_map, so_mcm, so_spectra, so_window, sph_tools


class TimeLogger:
    def __init__(self, verbose):
        self.t0 = None
        self.verbose = verbose

    def start(self, msg):
        if self.verbose:
            self.t0 = time.time()
            print(msg, end="... ", flush=True)

    def stop(self, msg="done"):
        if self.verbose:
            print(f"{msg} in {time.time() - self.t0:.2f} sec")


data_dir = os.path.dirname(__file__)
with open(os.path.join(data_dir, "parameters_test_data.yml"), "r") as stream:
    config = yaml.load(stream, Loader=yaml.FullLoader)


def do_simulation(car, window_data=None, verbose=True):
    time_log = TimeLogger(verbose)
    tmpl_config = config.get("car" if car else "healpix")

    if verbose:
        print("{} template: {}".format("CAR" if car else "HEALPIX", tmpl_config))

    if car:
        template = so_map.car_template(ncomp=config.get("ncomp"), **tmpl_config)
        binary = so_map.car_template(ncomp=1, **tmpl_config)
        # binary.data[:] = 0
        binary.data[1:-1, 1:-1] = 1
    else:
        nside = tmpl_config.get("nside")
        template = so_map.healpix_template(ncomp=config.get("ncomp"), nside=nside)
        binary = so_map.healpix_template(ncomp=1, nside=nside)
        vec = hp.ang2vec(tmpl_config.get("lon"), tmpl_config.get("lat"), lonlat=True)
        disc = hp.query_disc(nside, vec, radius=tmpl_config.get("radius") * np.pi / 180)
        binary.data[disc] = 1

    time_log.start("Generate binning file")
    binning_file = os.path.join(tempfile.gettempdir(), "binning.dat")
    pspy_utils.create_binning_file(bin_size=40, n_bins=100, file_name=binning_file)
    time_log.stop()

    time_log.start("Generate CMB")
    np.random.seed(config.get("seed"))
    cmb = template.synfast(os.path.join(data_dir, config.get("cl_file")))
    nsplits = config.get("nsplits", 2)
    rms_uKarcmin_T = config.get("rms_uKarcmin_T", 20)
    splits = [cmb.copy() for i in range(nsplits)]
    for i in range(nsplits):
        noise = so_map.white_noise(
            cmb, rms_uKarcmin_T=rms_uKarcmin_T, rms_uKarcmin_pol=np.sqrt(2) * rms_uKarcmin_T
        )
        splits[i].data += noise.data
    time_log.stop()

    if window_data is not None:
        window = binary.copy()
        if car:
            from pixell.enmap import ndmap

            window.data = ndmap(window_data, binary.data.wcs)
        else:
            window.data = window_data
    else:
        time_log.start("Generate mask")
        window = so_window.create_apodization(
            binary,
            apo_type="Rectangle" if car else "C1",
            apo_radius_degree=config.get("apo_radius_degree_survey"),
        )
        mask = so_map.simulate_source_mask(
            binary,
            n_holes=config.get("source_mask_nholes"),
            hole_radius_arcmin=config.get("source_mask_radius"),
        )
        mask = so_window.create_apodization(
            mask, apo_type="C1", apo_radius_degree=config.get("apo_radius_degree_mask")
        )
        window.data *= mask.data
        time_log.stop()

    lmax = config.get("lmax")
    niter = config.get("niter")
    time_log.start(f"Compute MCM and Bbl for lmax={lmax} and niter={niter}")
    windows = (window, window)
    mbb_inv, bbl = so_mcm.mcm_and_bbl_spin0and2(windows, binning_file, lmax=lmax, niter=niter)
    alms = [sph_tools.get_alms(split, windows, lmax=lmax, niter=niter) for split in splits]
    time_log.stop()

    time_log.start("Compute power spectra")
    spectra = config.get("spectra")
    Db = {
        f"split{i1}xsplit{i2}": so_spectra.bin_spectra(
            *so_spectra.get_spectra(alm1, alm2, spectra=spectra),
            binning_file,
            lmax,
            type="Dl",
            mbb_inv=mbb_inv,
            spectra=spectra,
        )
        for (i1, alm1), (i2, alm2) in cwr(enumerate(alms), 2)
    }
    time_log.stop()

    # Reduce size of map by only keeping first 1000 values
    if car:
        cmb.data = cmb.data[:, :100, :100]
    else:
        cmb.data = cmb.data[:, :1000]
    return {
        "cmb": np.asarray(cmb.data),
        "window": np.asarray(window.data),
        "mbb_inv": mbb_inv,
        "bbl": bbl,
        "spectra": Db,
    }


def store_test_data():
    test_data_dict = {"config": config}
    test_data_dict["car"] = do_simulation(car=True)
    test_data_dict["healpix"] = do_simulation(car=False)

    test_data_file = config.get("test_data_file", os.path.join(data_dir, "test_data.pkl"))
    with open(test_data_file, "wb") as f:
        pickle.dump(test_data_dict, f)


if __name__ == "__main__":
    store_test_data()



"""Tests for `pspy.so_dict` object."""

import tempfile
import unittest

from pspy import so_dict


class SODictTests(unittest.TestCase):
    def test_getter(self):
        d = so_dict.so_dict()
        with self.assertRaises(ValueError):
            _ = d["foo"]
        with self.assertRaises(ValueError):
            _ = d.get("foo", raise_error=True)

        self.assertEqual(d.get("foo", 666), 666)

    def test_read_from_file(self):
        s = b"""
        foo = 666
        bar = "bar"
        """
        with tempfile.NamedTemporaryFile() as tmp:
            tmp.write(s)
            tmp.seek(0)
            d = so_dict.so_dict()
            d.read_from_file(tmp.name)
            self.assertTrue(d.get("foo") != None)
            self.assertTrue(d.get("bar") == "bar")


if __name__ == "__main__":
    unittest.main()



from itertools import combinations_with_replacement as cwr
from pspy import so_spectra, so_cov
import matplotlib.pyplot as plt
import numpy as np

def append_spectra_and_cov(ps_dict, cov_dict, spec_list):
    """
    Get the power spectra vector containing
    the spectra in `spec_list` with the
    associated covariance matrix.

    Parameters
    ----------
    ps_dict: dict
      dict containing the different power spectra
    cov_list: dict
      dict containing the covariances
    spec_list: list
      list containing the name of the spectra

    """
    n_bins = len(ps_dict[spec_list[0]])
    n_spec = len(spec_list)

    spectra_vec = np.zeros(n_bins * n_spec)
    full_cov = np.zeros((n_bins * n_spec,
                         n_bins * n_spec))

    for i, key1 in enumerate(spec_list):
        spectra_vec[i*n_bins:(i+1)*n_bins] = ps_dict[key1]
        for j, key2 in enumerate(spec_list):
            if j < i: continue
            full_cov[i*n_bins:(i+1)*n_bins,
                     j*n_bins:(j+1)*n_bins] = cov_dict[key1, key2]

    full_cov = np.triu(full_cov)
    transpose_cov = full_cov.T
    full_cov += transpose_cov - np.diag(full_cov.diagonal())

    return spectra_vec, full_cov


def get_projector(n_bins, proj_pattern):
    """
    Get the projection operator
    to apply to a spectra vector to
    get the residual power spectra

    Parameters
    ----------
    n_bins: int
      number of bins
    proj_pattern: 1D array
      If the spectra vector is a concatenation
      of three spectra psA, psB, psC; and
      proj_pattern is [A, B, C], applying the
      projector to a spectra vector will give :
        A * psA + B * psB + C * psC
    """
    N = len(proj_pattern)
    identity = np.identity(n_bins)

    projector = (np.tile(identity, (1, N)) *
                 np.tile(np.repeat(proj_pattern, n_bins), (n_bins, 1)))
    return projector


def project_spectra_vec_and_cov(spectra_vec, full_cov, proj_pattern, calib_vec = None):
    """
    Get the residual spectrum and the associated
    covariance matrix from the spectra vector and
    the full covmat.

    Parameters
    ----------
    spectra_vec: 1D array
      spectra vector
    full_cov: 2D array
      full covariance matrix associated with spectra_vec
    proj_pattern: 1D array
      If the spectra vector is a concatenation
      of three spectra psA, psB, psC; and
      proj_pattern is [A, B, C], applying the
      projector to a spectra vector will give :
        A * psA + B * psB + C * psC
    calib_vec: 1D array
      calibration amplitudes to apply to the
      different power spectra
    """
    n_bins = len(spectra_vec) // len(proj_pattern)
    projector_uncal = get_projector(n_bins, proj_pattern)
    if calib_vec is None:
        calib_vec = np.ones(len(proj_pattern))
    projector_cal = get_projector(n_bins, proj_pattern * calib_vec)

    res_spectrum = projector_cal @ spectra_vec
    res_cov = projector_uncal @ full_cov @ projector_uncal.T

    return res_spectrum, res_cov

def get_chi2(spectra_vec, full_cov, proj_pattern, calib_vec = None, lrange = None):
    """
    Compute the chi2 of the residual
    power spectrum

    Parameters
    ----------
    spectra_vec: 1D array
      spectra vector
    full_cov: 2D array
      full covariance matrix associated with spectra_vec
    proj_pattern: 1D array (len = 3)
      If the spectra vector is a concatenation
      of three spectra psA, psB, psC; and
      proj_pattern is [A, B, C], applying the
      projector to a spectra vector will give :
        A * psA + B * psB + C * psC
    lrange: 1D array
    calib_vec: 1D array
      calibration amplitudes to apply to the
      different power spectra
    """
    res_spec, res_cov = project_spectra_vec_and_cov(spectra_vec, full_cov,
                                                    proj_pattern, calib_vec)
    if lrange is not None:
        return res_spec[lrange] @ np.linalg.inv(res_cov[np.ix_(lrange, lrange)]) @ res_spec[lrange]
    else:
        return res_spec @ np.linalg.inv(res_cov) @ res_spec

def plot_residual(lb, res_spec, res_cov, mode, title, file_name):
    """
    Plot the residual power spectrum and
    save it at a png file

    Parameters
    ----------
    lb: 1D array
    res_spec: 1D array
      Residual power spectrum
    res_cov: 2D array
      Residual covariance matrix
    mode: string
    title: string
    fileName: string
    """

    chi2 = res_spec @ np.linalg.inv(res_cov) @ res_spec

    plt.figure(figsize = (8, 6))
    plt.axhline(0, color = "k", ls = "--")
    plt.errorbar(lb, res_spec, yerr = np.sqrt(res_cov.diagonal()),
                 ls = "None", marker = ".",
                 label = f"Chi2 : {chi2:.1f}/{len(lb)}")
    plt.title(title)
    plt.xlabel(r"$\ell$")
    plt.ylabel(r"$\Delta D_\ell^\mathrm{%s}$" % mode)
    plt.tight_layout()
    plt.legend()
    plt.savefig(f"{file_name}.png", dpi = 300)

def get_calibration_amplitudes(spectra_vec, full_cov, proj_pattern, mode, lrange, chain_name):
    """
    Get the calibration amplitude and the
    associated error

    Parameters
    ----------
    spectra_vec: 1D array
    full_cov: 2D array
    proj_pattern: 1D array (len = 3)
      If the spectra vector is a concatenation
      of three spectra psA, psB, psC; and
      proj_pattern is [A, B, C], applying the
      projector to a spectra vector will give :
        A * psA + B * psB + C * psC
    mode: string
    lrange: 1D array
    chain_name: string
    """
    try:
        from cobaya.run import run
        from getdist.mcsamples import loadMCSamples
    except ModuleNotFoundError:
        raise ModuleNotFoundError("You need to install Cobaya to use this function")

    cal_vec = {# Global calibration :
               #  multiplicative factor to apply to
               # the different cross spectra to obtain
               # a calibration amplitude. Here the cross
               # spectra are [AxA, AxR, RxR] with A the
               # array you want to calibrate, and R the
               # reference array
              "TT": lambda c: np.array([c**2, c, 1]),
              # Pol. Eff.
              "EE": lambda e: np.array([e**2, e, 1]),
              "TE": lambda e: np.array([e, 1, 1]),
              "ET": lambda e: np.array([e, e, 1])}

    def logL(cal):
        if (proj_pattern == np.array([1, -1, 0])).all():
            if mode == "TT" or mode == "EE":
                # If we want to fit for the calib (resp. pol.eff.)
                # using the combination (c**2)AxA - c AxR,
                # we set the multiplicative factor to be
                # [c, 1, 1] such that we are minimizing
                # the residual c AxA - AxR
                calib_vec = np.array([cal, 1, 1])
        else:
            calib_vec = cal_vec[mode](cal)

        chi2 = get_chi2(spectra_vec, full_cov, proj_pattern, calib_vec, lrange)
        return -0.5 * chi2

    info = {
        "likelihood": {"my_like": logL},
        "params": {
            "cal": {
                "prior": {
                    "min": 0.5,
                    "max": 1.5
                         },
                "latex": "c"
                 }
                  },
        "sampler": {
            "mcmc": {
                "max_tries": 1e4,
                "Rminus1_stop": 0.001,
                "Rminus1_cl_stop": 0.03,
                    }
                   },
        "output": chain_name,
        "force": True
           }

    updated_info, sampler = run(info)
    samples = loadMCSamples(chain_name, settings = {"ignore_rows": 0.5})
    cal_mean = samples.mean("cal")
    cal_std = np.sqrt(samples.cov(["cal"])[0, 0])

    return cal_mean, cal_std



"""Tests for spectra generation."""
import os
import pickle
import unittest
from itertools import product

import numpy as np
from pspy.tests.data.generate_test_data import do_simulation

test_dir = os.path.dirname(__file__)
data_dir = os.path.join(test_dir, "data")


class SOSpectraTests(unittest.TestCase):
    def setUp(self, verbose=False):
        with open(os.path.join(data_dir, "test_data.pkl"), "rb") as f:
            self.reference = pickle.load(f)

        self.current = {
            kind: do_simulation(
                car=kind == "car", window_data=self.reference[kind]["window"], verbose=verbose
            )
            for kind in ["car", "healpix"]
        }

    def compare(self, ref, current, msg=""):
        self.assertIsInstance(ref, type(current), msg=msg)

        if isinstance(ref, dict):
            for k in ref.keys():
                self.compare(ref[k], current[k], msg=msg)
        elif isinstance(ref, np.ndarray):
            np.testing.assert_almost_equal(ref, current, err_msg=msg)
        else:
            self.assertTrue(True, f"Data type {type(ref)} are not compared!")

    def test_spectra(self, verbose=False):
        kinds = ["car", "healpix"]
        # do not test window function due to mismatch in distance computation
        # datas = ["cmb", "window", "mbb_inv", "bbl", "spectra"]
        datas = ["cmb", "mbb_inv", "bbl", "spectra"]
        for kind, data in product(kinds, datas):
            msg = f"Testing '{data}' data content ({kind} pixellisation)"
            if verbose:
                print(msg, end="... ", flush=True)
            self.compare(
                self.reference.get(kind).get(data), self.current.get(kind).get(data), msg=msg
            )
            if verbose:
                print("ok")


if __name__ == "__main__":
    unittest.main()



"""
Utils for pspy.
"""
import os

import numpy as np


def ps_lensed_theory_to_dict(filename, output_type, lmax=None, start_at_zero=False):
    """Read a lensed power spectrum from CAMB and return a dictionnary

    Parameters
    ----------
    filename : string
      the name of the CAMB lensed power spectrum you want to read
    lmax : integer
      the maximum multipole (spectra will be cut at)
    output_type :  string
      'Cl' or 'Dl'
    start_at_zero : boolean
      if True, ps start at l=0 and cl(l=0) and cl(l=1) are set to 0

    """

    fields = ["TT", "TE", "TB", "ET", "BT", "EE", "EB", "BE", "BB"]
    ps = {}
    l, ps["TT"], ps["EE"], ps["BB"], ps["TE"] = np.loadtxt(filename, unpack=True)
    ps["ET"] = ps["TE"].copy()
    ps["TB"], ps["BT"], ps["EB"], ps["BE"] = np.zeros((4, len(l)))

    if lmax is not None:
        l = l[:lmax]
    scale = l * (l + 1) / (2 * np.pi)
    for f in fields:
        if lmax is not None:
            ps[f] = ps[f][:lmax]
        if output_type == "Cl":
            ps[f] /= scale
        if start_at_zero:
            ps[f] = np.append(np.array([0, 0]), ps[f])
    if start_at_zero:
        l = np.append(np.array([0, 1]), l)
    return l, ps


def ps_from_params(cosmo_params, output_type, lmax, start_at_zero=False):

    """Given a set of cosmological parameters compute the corresponding lensed power spectrum
       You need to have camb installed to use this function
      ----------
      cosmo_params: dict
        dictionnary of cosmological parameters
        # e.g cosmo_params = {"cosmomc_theta":0.0104085, "logA": 3.044, "ombh2": 0.02237, "omch2": 0.1200, "ns": 0.9649, "Alens": 1.0, "tau": 0.0544}
      output_type :  string
        'Cl' or 'Dl'
      lmax: integer
        the maximum multipole to consider
      start_at_zero : boolean
        if True, ps start at l=0 and cl(l=0) and cl(l=1) are set to 0
        else, start at l=2
    """
    try:
        import camb
    except ModuleNotFoundError:
        raise ModuleNotFoundError("you need to install camb to use this function")

    if start_at_zero:
        lmin = 0
    else:
        lmin = 2

    camb_cosmo = {k: v for k, v in cosmo_params.items() if k not in ["logA", "As"]}
    camb_cosmo.update({"As": 1e-10*np.exp(cosmo_params["logA"]), "lmax": lmax, "lens_potential_accuracy": 1})
    pars = camb.set_params(**camb_cosmo)
    results = camb.get_results(pars)
    powers = results.get_cmb_power_spectra(pars, CMB_unit="muK", raw_cl=output_type == "Cl")
    l = np.arange(lmin, lmax)
    ps = {spec: powers["total"][l][:, count] for count, spec in enumerate(["TT", "EE", "BB", "TE" ])}
    ps["ET"] = ps["TE"].copy()
    for spec in ["TB", "BT", "EB", "BE" ]:
        ps[spec] = ps["TT"] * 0

    return l, ps


def get_nlth_dict(rms_uKarcmin_T, type, lmax, spectra=None, rms_uKarcmin_pol=None, beamfile=None):
    """Return the effective noise power spectrum Nl/bl^2 given a beam file and a noise rms

    Parameters
    ----------
    rms_uKarcmin_T: float
      the temperature noise rms in uK.arcmin
    type: string
      'Cl' or 'Dl'
    lmax: integer
      the maximum multipole to consider
    spectra: list of strings
      needed for spin0 and spin2 cross correlation, the arrangement of the spectra
    rms_uKarcmin_pol: float
      the polarisation noise rms in uK.arcmin
    beamfile: string
      the name of the beam transfer function (assuming it's given as a two column file l,bl)
    """

    if beamfile is not None:
        l, bl = np.loadtxt(beamfile, unpack=True)
    else:
        bl = np.ones(lmax + 2)

    l_th = np.arange(2, lmax + 2)
    nl_th = {}
    if spectra is None:
        nl_th["TT"] = (
            np.ones(lmax) * (rms_uKarcmin_T * np.pi / (60 * 180)) ** 2 / bl[2 : lmax + 2] ** 2
        )
        if type == "Dl":
            nl_th["TT"] *= l_th * (l_th + 1) / (2 * np.pi)
        return nl_th
    else:
        if rms_uKarcmin_pol is None:
            rms_uKarcmin_pol = rms_uKarcmin_T * np.sqrt(2)
        for spec in spectra:
            nl_th[spec] = np.zeros(lmax)
        nl_th["TT"] = np.ones(lmax) * (rms_uKarcmin_T * np.pi / (60 * 180)) ** 2 / bl[2 :lmax + 2] ** 2
        nl_th["EE"] = np.ones(lmax) * (rms_uKarcmin_pol * np.pi / (60 * 180)) ** 2 / bl[2 :lmax + 2] ** 2
        nl_th["BB"] = np.ones(lmax) * (rms_uKarcmin_pol * np.pi / (60 * 180)) ** 2 / bl[2 :lmax + 2] ** 2
        if type == "Dl":
            for spec in spectra:
                nl_th[spec] *= l_th * (l_th + 1) / (2 * np.pi)
    return l_th, nl_th


def read_beam_file(beamfile, lmax=None):
    """Read beam file with formal, l, bl, stuff and normalize it

    Parameters
    __________
    beamfile: string
      the name of the beam file
    lmax: integer
      the maximum multipole to consider
    """

    beam = np.loadtxt(beamfile)
    l, bl = beam[:, 0], beam[:, 1]
    if lmax is not None:
        l, bl = l[:lmax], bl[:lmax]

    return l, bl / bl[0]


def create_binning_file(bin_size, n_bins, lmax=None, file_name=None):
    """Create a (constant) binning file, and optionnaly write it to disk

    Parameters
    ----------
    bin_size: float
      the size of the bins
    n_bins: integer
      the number of  bins
    lmax: integer
      the maximum multipole to consider
    file_name: string
      the name of the binning file
    """
    bins = np.arange(n_bins)
    bin_low = bins * bin_size + 2
    bin_hi = (bins + 1) * bin_size + 1
    bin_cent = (bin_low + bin_hi) / 2

    if lmax is not None:
        id = np.where(bin_hi < lmax)
        bin_low, bin_hi, bin_cent = bin_low[id], bin_hi[id], bin_cent[id]

    if file_name is None:
        return bin_low, bin_hi, bin_cent
    else:
        f = open("%s" % file_name, mode="w")
        for i in range(len(bin_low)):
            f.write("%0.2f %0.2f %0.2f\n" % (bin_low[i], bin_hi[i], bin_cent[i]))
        f.close()


def read_binning_file(file_name, lmax):
    """Read a binningFile and truncate it to lmax, if bin_low lower than 2, set it to 2.
    format is bin_low, bin_high, bin_mean

    Parameters
    ----------
    binningfile: string
      the name of the binning file
    lmax: integer
      the maximum multipole to consider
    """
    bin_low, bin_hi, bin_cent = np.loadtxt(file_name, unpack=True)
    id = np.where(bin_hi < lmax)
    bin_low, bin_hi, bin_cent = bin_low[id], bin_hi[id], bin_cent[id]
    if bin_low[0] < 2:
        bin_low[0] = 2
    bin_hi = bin_hi.astype(int)
    bin_low = bin_low.astype(int)
    bin_size = bin_hi - bin_low + 1
    return bin_low, bin_hi, bin_cent, bin_size


def create_directory(name):
    """Create a directory

    Parameters
    ----------
    name: string
      the name of the directory
    """
    os.makedirs(name, exist_ok=True)


def naive_binning(l, fl, binning_file, lmax):
    """Bin a function of l given a binning file and lmax

    Parameters
    ----------
    l: 1d integer array
      the multipoles
    fl: 1d float array
      the 1-dimensional function to bin
    binning_file: string
      the name of the binning file
    lmax: integer
      the maximum multipole to consider

    """

    bin_low, bin_hi, bin_cent, bin_size = read_binning_file(binning_file, lmax)
    n_bins = len(bin_hi)
    fl_bin = np.zeros(len(bin_cent))
    for ibin in range(n_bins):
        loc = np.where((l >= bin_low[ibin]) & (l <= bin_hi[ibin]))
        fl_bin[ibin] = (fl[loc]).mean()
    return bin_cent, fl_bin


def beam_from_fwhm(fwhm_arcminute, lmax):
    """Compute the harmonic transform of the beam
    given the beam full width half maximum in arcminute

    Parameters
    ----------
    fwhm_arcminute: float
      full width half maximum in arcminute
    lmax: integer
      the maximum multipole to consider

    """
    beam_fwhm_rad = np.deg2rad(fwhm_arcminute) / 60
    fac = beam_fwhm_rad / np.sqrt(8 * np.log(2))

    ell = np.arange(2, lmax)
    bl = np.exp(-ell * (ell + 1) * fac ** 2 / 2.0)
    return ell, bl


def create_arbitrary_binning_file(delta_l_list, l_bound_list, binning_file=None):
    """
    This function return a binning scheme with non constant bin size
    you specify a list of bin size: delta_l_list, and the corresponding bound at
    which this bin size ceases to apply: l_bound_list
    optionnaly write to disk a binning file in format, lmin, lmax, lmean

    Parameters
    ----------
    delta_l_list: list of integers
        the different binsize we want to consider
    l_bound_list: list of integers
        the multipole at which the binsize ceases to apply
    write_binning_file: string
        if you want to write to disk, pass a string with the name of the file
    """
 
    lmin_list = []
    lmax_list = []
    lmean_list = []
    
    lmin = 2
    for count, ells in enumerate(l_bound_list):
        while True:
            lmax = lmin +  delta_l_list[count]

            if lmax > l_bound_list[count]:
                break

            lmean = (lmax + lmin) / 2

            lmin_list += [lmin]
            lmax_list += [lmax]
            lmean_list += [lmean]
            
            lmin = lmax + 1
    if binning_file is not None:
        np.savetxt(binning_file, np.transpose([lmin_list, lmax_list, lmean_list]))
    
    return lmin_list, lmax_list, lmean_list

def maximum_likelihood_combination(cov_mat, P_mat, data_vec, test_matrix=False):
    """
    This function solve for the maximum likelihood data_vec and covariance for a problem of the form
    chi2 = (data_vec - P ML_data_vec).T cov_mat ** -1 data_vec - P ML_data_vec)


    Parameters
    ----------
    cov_mat: 2d array
        the covariance matrix of data_vec, of size N_ini x N_ini
    P_mat: 2d array
        the "pointing matrix" that project the final data vector of size N_f into the initial data vector space N_ini
        shape is (N_ini, N_f)
    data_vec: 1d array
        the initial data vector
    """
    
    inv_cov_mat = np.linalg.inv(cov_mat)

    ML_cov_mat = np.linalg.inv(np.dot(np.dot(P_mat, inv_cov_mat), P_mat.T))
    ML_data_vec = np.dot(ML_cov_mat, np.dot(P_mat, np.dot(inv_cov_mat, data_vec)))
    
    if test_matrix:
        is_symmetric(ML_cov_mat, tol=1e-7)
        is_pos_def(ML_cov_mat)

    return ML_cov_mat, ML_data_vec


def is_symmetric(mat, tol=1e-8):
    """
    This function check if a matrix is symmetric
    
    Parameters
    ----------
    mat: 2d array
        the matrix tested
    tol:
        the absolute tolerance on assymetry
    """
    if np.all(np.abs(mat-mat.T) < tol):
        print("the tested matrix is symmetric")
    else:
        print("the tested matrix is not symmetric")
    return


def is_pos_def(mat):
    """
    This function check if a matrix is positive definite
    
    Parameters
    ----------
    mat: 2d array
        the matrix tested
    """
    if np.all(np.linalg.eigvals(mat) > 0):
        print("the tested matrix is positive definite")
    else:
        print("the tested matrix is not positive definite")
        print(np.linalg.eigvals(mat))
    return


    



import setuptools
from numpy.distutils.core import setup
from numpy.distutils.extension import Extension
from setuptools import find_packages

import versioneer

with open("README.rst") as readme_file:
    readme = readme_file.read()

compile_opts = {
    "extra_f90_compile_args": [
        "-fopenmp",
        "-ffree-line-length-none",
        "-fdiagnostics-color=always",
        "-Wno-tabs",
    ],
    "f2py_options": ["skip:", "map_border", "calc_weights", ":"],
    "extra_link_args": ["-fopenmp"],
}

mcm = Extension(
    name="pspy.mcm_fortran.mcm_fortran",
    sources=["pspy/mcm_fortran/mcm_fortran.f90", "pspy/wigner3j/wigner3j_sub.f"],
    **compile_opts
)
cov = Extension(
    name="pspy.cov_fortran.cov_fortran",
    sources=["pspy/cov_fortran/cov_fortran.f90", "pspy/wigner3j/wigner3j_sub.f"],
    **compile_opts
)

setup(
    name="pspy",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Simons Observatory Collaboration Power Spectrum Task Force",
    url="https://github.com/simonsobs/pspy",
    description="Python power spectrum code",
    long_description=readme,
    long_description_content_type="text/x-rst",
    license="BSD license",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    entry_points={},
    ext_modules=[mcm, cov],
    install_requires=[
        "numpy>=1.20",
        "healpy",
        "pyFFTW",
        "pillow",  # this one should be installed by pixell
        "pixell>=0.7.0",
    ],
    packages=find_packages(),
    package_data={"pspy": ["pspy/tests/data/*.pkl"]},
    include_package_data=True,
    scripts=["scripts/test-pspy"],
)
